/****************************************************************
 * See file COPYING included in distribution for license details
 ****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <values.h>
#include <math.h>
#include <assert.h>

#include "boxmap.h"
#include "db-interfaces/postgres/vrml-postgres.h"

#define X_FIELD 0
#define Y_FIELD 1
#define Z_FIELD 2

/*************  DBFieldLocation_new  *************/
DBFieldLocation*
DBFieldLocation_new (const char* value, float position )
{
   
   DBFieldLocation *p = (DBFieldLocation*)malloc(sizeof(DBFieldLocation));
   p->value = strdup(value);
   p->position = position;

   return p;
}

/***********  DBFieldLocation_delete  **************/
void
DBFieldLocation_delete (DBFieldLocation *p)
{
   if ( p ) {
      if ( p->value )
         free (p->value);
      free (p);
   }
}

/****************  BoxMap_new  *******************/
BoxMap*
BoxMap_new()
{
   BoxMap *bm = (BoxMap*)malloc( sizeof(BoxMap) );
   bm->x_axis = bm->y_axis = bm->z_axis = NULL;
   return bm;
}


/***************  BoxMap_delte  ******************/
void
BoxMap_delete( BoxMap* bm)
{
   if ( bm ) {
      CoordMap_delete (bm->x_axis);
      CoordMap_delete (bm->y_axis);
      CoordMap_delete (bm->z_axis);
      free (bm);
   }
}

/********  BoxMap_replace_CoordMap_N  **************/
void
BoxMap_replace_CoordMap_N      (BoxMap* bm, CoordMap* c, int idx )
{
   if ( idx == 0 ) {
      CoordMap_delete (bm->x_axis);
      bm->x_axis = c;
   }
   else if ( idx == 1 ) {
      CoordMap_delete (bm->y_axis);
      bm->y_axis = c;
   }
   else if ( idx == 2 ) {
      CoordMap_delete (bm->z_axis);
      bm->z_axis = c;
   }
   
}

/**************  BoxGraphData_new  *****************/
BoxGraphData*
BoxGraphData_new()
{
   BoxGraphData *g   = (BoxGraphData*)malloc( sizeof(BoxGraphData) );
   g->type           = BoxMap_1;
   g->box_map        = BoxMap_new();
   g->fields         = NULL;

   g->appearance.text_justify_mode = TEXT_JUSTIFY_FONTSTYLE;
   
   SFColor_set( &g->appearance.box_color, 0, 0.5, 0.2 );
   
   SimpleGraphItem_set ( &g->appearance.x_axis_labels,
                         1,
                         0.1, 0.23, 0.38);
   
   SimpleGraphItem_set ( &g->appearance.x_axis_heading,
                         0,
                         0.1, 1.0, 0.1 );
   
   SimpleGraphItem_set ( &g->appearance.y_axis_values,
                         1,
                         0.1, 0.2, 0.65);

   SimpleGraphItem_set ( &g->appearance.graph_lines,
                         1,
                         0.8, 0.0, 0.2 );

   SimpleGraphItem_set ( &g->appearance.z_axis_labels,
                         1,
                         0.1, 0.23, 0.38);

   SimpleGraphItem_set ( &g->appearance.z_axis_heading,
                         0,
                         0.1, 1.0, 0.1 );
   
   SimpleGraphItem_set ( &g->appearance.elevation_grid,
                         1,
                         0.1, 0.4, 0.5 );



   return g;
}

/***************  BoxGraphData_delete  *****************/
void
BoxGraphData_delete (BoxGraphData* g)
{
   if ( g ) {
      BoxMap_delete (g->box_map);
      free (g);
   }
   
}

/*********  BoxMapGraphData_replace_BoxMap  ***********/
void
BoxGraphData_replace_BoxMap (BoxGraphData* d, BoxMap* bm)
{
   if ( d ) {
      BoxMap_delete (d->box_map);
      d->box_map = bm;
   }
}


/***************  location_set  **********************/
int
location_set (DBFieldOrder *fld, const char *test )
{
   int i;
   DBFieldLocation *loc;
   
   for ( i=0; i < fld->n_axis; i++ ) {
      loc = fld->axis[i];
      if ( strcasecmp( loc->value, test) == 0 )
         return 1;
   }
   return 0;
}

/***********  strip_trailing_spaces  *****************/
char *
strip_trailing_spaces (const char *str)
{
   static char strip_str[1024];
   
   int i;

   strip_str[0] = 0;
   
   if ( str ) {
      strncpy( strip_str, str, 1024 );
      i = strlen(strip_str) -1;
      /*      printf ( "%s len is %d\n", strip_str, i ); */
      while ( i > 0 ) {
         if ( strip_str[i] == ' ' )
            strip_str[i] = 0;
         else
            break;
         i --;
      }
   }
   return strip_str;
}

/****************  location_value  *******************/
float
location_value (DBFieldOrder *fld, const char *test)
{
   int i;
   DBFieldLocation *loc;
   
   for ( i=0; i < fld->n_axis; i++ ) {
      loc = fld->axis[i];
      if ( strcasecmp( loc->value, test ) == 0 )
         return loc->position;
   }
   /* Error! */
   return MINFLOAT;
}

/****************  get_min_z  ************************/
int
get_min_max_y ( DATA_ACCESS_FNXS *fnxs, BoxGraphData *g_data, float *min, float *max)
{
   DBFieldOrder       *field;
   int                n_tuples, i;
   char               err_msg[256];
   float              test;
   
   strcpy (err_msg, "get_min_max_y: " );
   field = g_data->fields[Y_FIELD];
   
   *min = MAXFLOAT;
   *max = MINFLOAT;

   n_tuples = (*fnxs->num_tuples_fnx)(fnxs->instance_data);
   
   for ( i=0; i < n_tuples; i++ ) {
      if ( ! (*fnxs->is_null_fnx)( fnxs->instance_data, i, field->field_order ) ) {
         test = atof( (*fnxs->get_value_fnx)( fnxs->instance_data, i, field->field_order ) );
         if ( test < *min )
            *min = test;
         if ( test > *max )
            *max = test;
      }
      
   }
   
   return 0;

}

/**************  get_units_x_z  **********************/
int
get_units_x_z (DATA_ACCESS_FNXS *fnxs,BoxGraphData *g_data, int *n_x_units, int *n_y_units )
{
   DBFieldOrder       *x_field, *z_field;
   int                n_tuples, i, j;
   char               **x_found = NULL, **z_found = NULL;
   int                x_count = 0, z_count = 0;
   char               test[512];
   
/*     Pretty stupid algorithm that is horribly inefficient but should */
/*     serve our purposes for small samples */
   x_field = g_data->fields[X_FIELD];
   z_field = g_data->fields[Z_FIELD];
   
   n_tuples = (*fnxs->num_tuples_fnx)( fnxs->instance_data );
   memset (test, 0x00, 512 );
   
   for ( i=0; i < n_tuples; i++ ) {
      strncpy( test, (*fnxs->get_value_fnx)( fnxs->instance_data, i, x_field->field_order), 500 );
      for ( j=0; j < x_count; j++ ) {
         if ( strcasecmp( test, x_found[j] ) == 0 )
            break;
      }
      if ( j == x_count ) {
         /* did not find it */
         if ( (x_found = (char**)realloc( x_found, sizeof(char*)*x_count + 1)) == NULL ) {
            fprintf ( stderr, "Out of memory count X units\n" );
            exit (1);
         }
         x_found[x_count] = strdup( test );
         x_count ++;
      }

      /*  Look for Z units */
      strncpy( test, (*fnxs->get_value_fnx)( fnxs->instance_data, i, z_field->field_order), 500 );
      for ( j=0; j < z_count; j++ ) {
         if ( strcasecmp( test, z_found[j] ) == 0 )
            break;
      }
      if ( j == z_count ) {
         /* did not find it */
         if ( (z_found = (char**)realloc( z_found, sizeof(char*)*z_count + 1)) == NULL ) {
            fprintf ( stderr, "Out of memory count Z units\n" );
            exit (1);
         }
         z_found[z_count] = strdup( test );
         z_count ++;
      }
   }

   *n_x_units = x_count;
   *n_y_units = z_count;

   if ( x_found ) {
      for ( j=0; j < x_count; j++ ) {
         free (x_found[j]);
      }
      free (x_found);
   }

   if ( z_found ) {
      for ( j=0; j < z_count; j++ ) {
         free (z_found[j]);
      }
      free (z_found);
   }

   
   return 0;
}

/*************  get_x_z_field_axis  ******************/
int
get_x_z_field_axis (DATA_ACCESS_FNXS *fnxs, BoxGraphData *g_data, int box_width, int box_depth )
{
   int   rc;
   int   n_x_units, n_z_units;
   float x_pos, z_pos;
   float x_incr,  z_incr;
   char  value[512];
   int   tuples;
   int   i;
   DBFieldOrder **f_list;
   
   DBFieldOrder *field;
   
   if ( (rc = get_units_x_z( fnxs, g_data, &n_x_units, &n_z_units)) != 0)
      return rc;

   x_incr = ((float)box_width) + ((float)box_width * 1.8 );
   z_incr = ((float)box_depth) + ((float)box_depth * 1.2);
   
   /* Offset form 0 origin */
   x_pos = -((n_x_units / 2) * x_incr);
   z_pos = 0; /*  -((n_z_units / 2) * z_incr); */

   f_list = g_data->fields;
   tuples = (*fnxs->num_tuples_fnx)( fnxs->instance_data );
   for ( i=0 ; i < tuples; i++ ) {
      // Test for X field location
      field = f_list[X_FIELD];
      strcpy ( value, (*fnxs->get_value_fnx)(fnxs->instance_data, i, field->field_order ));
      if ( !location_set (field, value) ) {
         field->axis = (DBFieldLocation**)realloc( field->axis, (field->n_axis+1) * sizeof(DBFieldLocation*));
         if ( field->axis == NULL ) {
            fprintf (stderr, "Out of memory allocating FieldLocations !\n");
            exit (1);
         }
         field->axis[ field->n_axis ] = DBFieldLocation_new( value, x_pos );
         field->n_axis ++;
         x_pos += x_incr;
      }

      // Test for Z field location
      field = f_list[Z_FIELD];
      strcpy ( value, (*fnxs->get_value_fnx)( fnxs->instance_data, i, field->field_order ));
      if ( !location_set (field, value) ) {
         field->axis = (DBFieldLocation**)realloc( field->axis, (field->n_axis+1) * sizeof(DBFieldLocation*));
         if ( field->axis == NULL ) {
            fprintf (stderr, "Out of memory allocating FieldLocations !\n");
            exit (1);
         }
         field->axis[ field->n_axis ] = DBFieldLocation_new( value, z_pos );
         field->n_axis ++;
         z_pos += z_incr;
      }
      
   }
   
   return 0;
}

/**************  postgres_graph_item_box  **********************/
void
graph_item_box (FILE *fp, DATA_ACCESS_FNXS *fnxs,BoxGraphData* g_data,
                int tuple, float scale_y,  float min_y, int box_width, int box_depth, float *max_z  )
{
   float x_pos, y_pos, z_pos;
   float y_scaled_pos;
   SFColor       *color;
   
   DBFieldOrder  *x_field, *y_field, *z_field;

   x_field = g_data->fields[X_FIELD];
   y_field = g_data->fields[Y_FIELD];
   z_field = g_data->fields[Z_FIELD];

   // Don't graph it if any value is NULL
   if ( (*fnxs->is_null_fnx)( fnxs->instance_data, tuple, x_field->field_order ) ||
        (*fnxs->is_null_fnx)( fnxs->instance_data, tuple, y_field->field_order ) ||
        (*fnxs->is_null_fnx)( fnxs->instance_data, tuple, z_field->field_order ) )
      return;

   x_pos = location_value( x_field, (*fnxs->get_value_fnx)(fnxs->instance_data, tuple, x_field->field_order ));
   if ( x_pos == MINFLOAT )
      fprintf (stderr, "Error finding X value %s,continuing...\n", x_field->field_name );

   z_pos = location_value( z_field, (*fnxs->get_value_fnx)(fnxs->instance_data, tuple,  z_field->field_order ));
   if ( z_pos == MINFLOAT )
      fprintf (stderr, "Error finding Z value %s,continuing...\n", z_field->field_name );

   y_pos = atof( (*fnxs->get_value_fnx)( fnxs->instance_data, tuple, y_field->field_order ));

   y_scaled_pos = (y_pos - min_y) / scale_y;

   color = &g_data->appearance.box_color;
   
   fprintf ( fp, "\tTransform {\n" );
   fprintf ( fp, "\t\t translation %f %f %f\n", (float)x_pos, (float)(y_scaled_pos / 2.0), (float)z_pos + 1.0);
   fprintf ( fp, "\t\t children [\n");
   fprintf ( fp, "\t\t\t Shape {\n");
   fprintf ( fp, "\t\t\t\t geometry Box { size %d %f %d }\n",
                           (int)box_width, (float)y_scaled_pos, (int)box_depth );
   fprintf ( fp, "\t\t\t\t appearance Appearance {\n" );
   fprintf ( fp, "\t\t\t\t\t material Material { diffuseColor %1.2f %1.2f %1.2f }\n", color->R, color->G, color->B );
   fprintf ( fp, "\t\t\t\t }\n");
   fprintf ( fp, "\t\t\t }\n" );
   fprintf ( fp, "\t\t ]\n");
   fprintf ( fp, "\t}\n\n");

   if ( z_pos > *max_z ) {
      *max_z = z_pos;
   }
   
   
}
/******************  add_frame  ***************************/
void
add_frame (FILE* fp, DATA_ACCESS_FNXS *fnxs, BoxGraphData *g_data, float min_y, float max_y, float z_pos, int box_width )
{
   #define N_Y_LABELS   5
   
   BoxMap               *box_map      = g_data->box_map;
   DBFieldOrder         **f_list      = g_data->fields;
   DBFieldOrder         *field;
   CoordMap             *coord_map;
   char                 *tmp_char_ptr;
   //   float                x_axis;
   //   float                z_axis;
   int                  i, j;
   float                min_z;
   float                min_x, max_x;
   float                y_incr, incr_value;
   float                y_label_pos[N_Y_LABELS];
   float                y_label_value[N_Y_LABELS];
   float                axis_z_frame;
   char                 str_field_value[256];
   int                  str_field_len;
   SFColor              *color;
   
   field = f_list[X_FIELD];
   if ( field->n_axis <= 0 )
      return;
   min_x = field->axis[0]->position - box_width;

   field = f_list[Z_FIELD];
   if ( field->n_axis <= 0 )
      return;
   min_z = (field->axis[field->n_axis-1]->position) + 25;  // should be + box_depth + 2
   
   // Add x labels
   field = f_list[X_FIELD];
   if ( g_data->appearance.x_axis_labels.show ) {
      color = &g_data->appearance.x_axis_labels.color;
      for (i=0; i < field->n_axis; i++ ) {
         fprintf ( fp, "\tTransform {\n" );
         fprintf ( fp, "\t\ttranslation %f %f %f\n", (float)field->axis[i]->position, (float)0, (float)min_z );
         fprintf ( fp, "\t\tscale 20 20 20\n");
         fprintf ( fp, "\t\trotation 1 0 0 -1.5707\n" );
         fprintf ( fp, "\t\tchildren [\n");
         fprintf ( fp, "\t\t\tTransform {\n" );
         fprintf ( fp, "\t\t\t\trotation 0 0 1 -1.5707\n");
         fprintf ( fp, "\t\t\t\tchildren [\n");
         fprintf ( fp, "\t\t\t\t\tShape {\n");
         fprintf ( fp, "\t\t\t\t\t\tgeometry Text { string \"%s\" }\n", field->axis[i]->value );
         fprintf ( fp, "\t\t\t\t\t\t\tappearance Appearance {\n" );
         fprintf ( fp, "\t\t\t\t\t\t\t\tmaterial Material { diffuseColor %1.2f %1.2f %1.2f }\n",
                   color->R, color->G, color->B );
         fprintf ( fp, "\t\t\t\t\t\t\t}\n");     // Appearance
         fprintf ( fp, "\t\t\t\t\t\t}\n" );   // Shape
         fprintf ( fp, "\t\t\t\t\t]\n");    // Children
         fprintf ( fp, "\t\t\t\t}\n" );    // Transform
         fprintf ( fp, "\t\t\t]\n" );    // children
         fprintf ( fp, "\t}\n\n");
      }
   }      // If show x-axis-labels
   
   // X heading label
   
   coord_map = box_map->x_axis;
   tmp_char_ptr = coord_map->graph_name != NULL ? coord_map->graph_name : coord_map->column_name;
   // Sorta centered
   str_field_len = (strlen( tmp_char_ptr ) / 2) * 7;

   if ( g_data->appearance.x_axis_heading.show ) {
      color = &g_data->appearance.x_axis_heading.color;
      fprintf ( fp, "\tTransform {\n" );
      fprintf ( fp, "\t\tscale 20 20 20\n" );
      fprintf ( fp, "\t\ttranslation -%f 0 %f\n", (float)(str_field_len),  (float)(min_z) );
      fprintf ( fp, "\t\t\tchildren [\n" );
      fprintf ( fp, "\t\t\t\tShape {\n" );
      fprintf ( fp, "\t\t\t\t\tgeometry Text { string \"%s\" }\n", tmp_char_ptr );
      fprintf ( fp, "\t\t\t\t\tappearance Appearance {\n" );
      fprintf ( fp, "\t\t\t\t\t\tmaterial Material { diffuseColor %1.2f %1.2f %1.2f }\n", color->R, color->G, color->B );
      fprintf ( fp, "\t\t\t\t\t}\n" );       // appearance
      fprintf ( fp, "\t\t\t\t}\n" );       // shape
      fprintf ( fp, "\t\t\t]\n" );       // end children
      fprintf ( fp, "\t}\n" );         // transform
   }
   
   // setup some axis variable
   axis_z_frame = min_z + box_width;
   max_x = box_width - (min_x*2);

   /* Z labels */
   field = f_list[Z_FIELD];
   if ( g_data->appearance.z_axis_labels.show ) {
      color = &g_data->appearance.z_axis_labels.color;
      for (i=0; i < field->n_axis; i++ ) {
         tmp_char_ptr = strip_trailing_spaces (field->axis[i]->value);

/*       Should be able to simply justify the text with a FontStyle but not all */
/*       VRML viewers support this when scaling.  This method is a bit simpler and */
/*       should work for larger number of viewers (IF scaling 20 this seems to work OK) */
         if ( g_data->appearance.text_justify_mode == TEXT_JUSTIFY_ESTIMATE_1 ) {
            str_field_len = strlen (tmp_char_ptr ) * 8;            
         }
         else {
            assert ( g_data->appearance.text_justify_mode == TEXT_JUSTIFY_FONTSTYLE );
            str_field_len = 0;
         }         
         fprintf ( fp, "\tTransform {\n" );
         fprintf ( fp, "\t\ttranslation %f 0 %f\n",
                           (float)(min_x-str_field_len), (float)field->axis[i]->position+ (box_width/2) );
         fprintf ( fp, "\t\tscale 20 20 20\n");
         fprintf ( fp, "\t\trotation 1 0 0 -1.5707\n" );
         fprintf ( fp, "\t\tchildren [\n");
         fprintf ( fp, "\t\t\tTransform {\n" );
         fprintf ( fp, "\t\t\t\tchildren [\n");
         fprintf ( fp, "\t\t\t\t\tShape {\n");
         fprintf ( fp, "\t\t\t\t\t\tgeometry Text { \n" );
         fprintf ( fp, "\t\t\t\t\t\t\tstring \"%s\" \n", tmp_char_ptr );
         if ( g_data->appearance.text_justify_mode == TEXT_JUSTIFY_FONTSTYLE ) {
            fprintf ( fp, "\t\t\t\t\t\t\tfontStyle FontStyle {\n" );
            fprintf ( fp, "\t\t\t\t\t\t\t\tjustify \"END\"\n" ); 
            fprintf ( fp, "\t\t\t\t\t\t\t}\n" ); 
         }
         fprintf ( fp, "\t\t\t\t\t\t}\n");
         fprintf ( fp, "\t\t\t\t\t\t\tappearance Appearance {\n" );
         fprintf ( fp, "\t\t\t\t\t\t\t\tmaterial Material { diffuseColor %1.2f %1.2f %1.2f }\n",
                   color->R, color->G, color->B );
         fprintf ( fp, "\t\t\t\t\t\t\t}\n");     // Appearance
         fprintf ( fp, "\t\t\t\t\t\t}\n" );   // Shape
         fprintf ( fp, "\t\t\t\t\t]\n");    // Children
         fprintf ( fp, "\t\t\t\t}\n" );    // Transform
         fprintf ( fp, "\t\t\t]\n" );    // children
         fprintf ( fp, "\t}\n\n");
      }
   }      // If show x-axis-labels
   
   // Add y axis
   color = &g_data->appearance.graph_lines.color;
   fprintf ( fp, "\tTransform {\n" );
   fprintf ( fp, "\t\t translation %f %f %f\n", (float)min_x, (float)0, (float)min_z );
   fprintf ( fp, "\t\t children [\n");
   fprintf ( fp, "\t\t\t Transform {\n" );
   fprintf ( fp, "\t\t\t\t children [\n");
   fprintf ( fp, "\t\t\t\t\t Shape {\n");
   fprintf ( fp, "\t\t\t\t\t\t geometry IndexedLineSet { \n" );
   fprintf ( fp, "\t\t\t\t\t\t\t colorPerVertex FALSE \n" );
   fprintf ( fp, "\t\t\t\t\t\t\t coord Coordinate { point[ %f 0 0, %f 100 0, 0 0 -%f, 0 100 -%f, %f 0 -%f, %f 100 -%f] }\n",
             (float)max_x,
             (float)max_x,
             (float)axis_z_frame,
             (float)axis_z_frame,
             (float)max_x, (float)axis_z_frame,
             (float)max_x, (float)axis_z_frame );
   fprintf ( fp, "\t\t\t\t\t\t\t coordIndex [ 0 1 -1 2 3 -1 4 5] \n" );
   fprintf ( fp, "\t\t\t\t\t\t\t\t color Color { color [ 0.8 0 0, 0.8 0 0, 0.8 0 0 ]}\n" );
   fprintf ( fp, "\t\t\t\t\t\t\t }\n");     // Color
   fprintf ( fp, "\t\t\t\t\t\t }\n" );   // Shape
   fprintf ( fp, "\t\t\t\t\t ]\n");    // Children
   fprintf ( fp, "\t\t\t\t }\n" );    // Transform
   fprintf ( fp, "\t\t\t ]\n" );    // children
   fprintf ( fp, "\t}\n\n");
   
   // Add y labels
   y_incr = (float)((fabs( max_y - min_y)) / (float)N_Y_LABELS);
   incr_value = 0;
   for (i=0; i < N_Y_LABELS; i++ ) {
      y_label_value[i] = min_y;
      min_y     += y_incr;

      y_label_pos[i] = incr_value;
      incr_value     += 100 / N_Y_LABELS;
   }

   fprintf ( fp, "\tTransform {\n" );
   fprintf ( fp, "\t\t translation %f %f %f\n", (float)min_x, (float)0, (float)min_z );
   fprintf ( fp, "\t\t children [\n");
   fprintf ( fp, "\t\t\t Transform {\n" );
   fprintf ( fp, "\t\t\t\t children [\n");
   fprintf ( fp, "\t\t\t\t\t Shape {\n");
   fprintf ( fp, "\t\t\t\t\t\t geometry IndexedLineSet { \n" );
   fprintf ( fp, "\t\t\t\t\t\t\t colorPerVertex FALSE \n" );
   fprintf ( fp, "\t\t\t\t\t\t\t coord Coordinate { point [ " );
   // Horizontal lines labeling Y values - triplets of points
   for ( i=0; i < N_Y_LABELS-1; i++ ) {
      fprintf ( fp, "%f %f 0, %f %f -%f, 0 %f -%f, ",
                (float)max_x, (float)y_label_pos[i],  /* 0 */
                (float)max_x, (float)y_label_pos[i],  (float)axis_z_frame,
                /* 0 */       (float)y_label_pos[i],  (float)axis_z_frame );
   }
   fprintf ( fp, "%f %f 0, %f %f -%f, 0 %f -%f ] }\n",
             (float)max_x, (float)y_label_pos[i], /* 0 */
             (float)max_x, (float)y_label_pos[i], (float)axis_z_frame,
             /* 0 */       (float)y_label_pos[i], (float)axis_z_frame );
   // Coordinate indices
   fprintf ( fp, "\t\t\t\t\t\t\t coordIndex [ " );
   for ( i=0, j=0; i < N_Y_LABELS; i++ ) {
      fprintf ( fp, "%d %d %d -1 ", j, j+1, j+2 );
      j += 3;
   }
   fprintf ( fp, "] \n" );
   // Color Nodes
   fprintf ( fp, "\t\t\t\t\t\t\t\t color Color { color [ " );
   for ( i=0; i < N_Y_LABELS-1; i ++ ) {
      fprintf (fp, "0.8 0.1 0, ");
   }
   fprintf ( fp, " 0.8 0.1 0 ] } \n" );
   fprintf ( fp, "\t\t\t\t\t\t\t } \n");     // Color
   fprintf ( fp, "\t\t\t\t\t\t }\n" );   // Shape

   // Y text values
   if ( g_data->appearance.y_axis_values.show ) {
      color = &g_data->appearance.y_axis_values.color;

      for ( i=0; i < N_Y_LABELS; i++ ) {
         str_field_len = sprintf( str_field_value, "%f", (float)y_label_value[i] );
         // Varying support in VRML viewers.  This estimate seems to work OK for
         // some (at least libVRML97) which don't support FontStyle justify on scaled
         // text
         if ( g_data->appearance.text_justify_mode == TEXT_JUSTIFY_ESTIMATE_1 ) {
            str_field_len = -(str_field_len * 6);
         }
         else {
            assert (g_data->appearance.text_justify_mode == TEXT_JUSTIFY_FONTSTYLE );
            str_field_len = min_x;
         }
         fprintf ( fp, "\t\t\t\t\t Transform {\n" );
         fprintf ( fp, "\t\t\t\t\t\t scale 15 15 15\n" );
         fprintf ( fp, "\t\t\t\t\t\t translation  %d %f -%f\n", str_field_len, (float)y_label_pos[i],
                   (float)axis_z_frame );
         fprintf ( fp, "\t\t\t\t\t\t children [\n" );
         fprintf ( fp, "\t\t\t\t\t\t\t Shape {\n" );
         fprintf ( fp, "\t\t\t\t\t\t\t\t geometry Text {\n" );
         fprintf ( fp, "\t\t\t\t\t\t\t\t\t string \"%s\"\n", str_field_value );
         if ( g_data->appearance.text_justify_mode == TEXT_JUSTIFY_FONTSTYLE ) {
            fprintf ( fp, "\t\t\t\t\t\t\tfontStyle FontStyle {\n" );
            fprintf ( fp, "\t\t\t\t\t\t\t\tjustify \"END\"\n" ); 
            fprintf ( fp, "\t\t\t\t\t\t\t}\n" ); 
         }
         fprintf ( fp, "\t\t\t\t\t\t\t\t }\n" );
         fprintf ( fp, "\t\t\t\t\t\t\t\t appearance Appearance {\n" );
         fprintf ( fp, "\t\t\t\t\t\t\t\t\t material Material { diffuseColor %1.2f %1.2f %1.2f }\n",
                   color->R, color->G, color->B );
         fprintf ( fp, "\t\t\t\t\t\t\t\t }\n");
         fprintf ( fp, "\t\t\t\t\t\t\t }\n");       // Shape
         fprintf ( fp, "\t\t\t\t\t\t ]\n" );    // children
         fprintf ( fp, "\t\t\t\t\t }\n" );    // Transform
      }
   
      fprintf ( fp, "\t\t\t\t\t ]\n");    // Children
      fprintf ( fp, "\t\t\t\t }\n" );    // Transform
      fprintf ( fp, "\t\t\t ]\n" );    // children
      fprintf ( fp, "\t}\n\n");
   }

   // Z heading label
   coord_map = box_map->y_axis;
   tmp_char_ptr = coord_map->graph_name != NULL ? coord_map->graph_name : coord_map->column_name;

   
   if ( g_data->appearance.z_axis_heading.show ) {
      color = &g_data->appearance.z_axis_heading.color;
      fprintf ( fp, "\tTransform {\n" );
      fprintf ( fp, "\t\tscale 20 20 20\n" );
      fprintf ( fp, "\t\ttranslation %f 0 %f\n", (float)(min_x - 10),  (float)(min_z) );
      fprintf ( fp, "\t\t\tchildren [\n" );
      fprintf ( fp, "\t\t\t\tShape {\n" );
      fprintf ( fp, "\t\t\t\t\tappearance Appearance {\n" );
      fprintf ( fp, "\t\t\t\t\t\tmaterial Material { diffuseColor %1.2f %1.2f %1.2f }\n",
                color->R, color->G, color->B );
      fprintf ( fp, "\t\t\t\t\t}\n" );       // appearance
      fprintf ( fp, "\t\t\t\t\tgeometry Text { \n" );
      fprintf ( fp, "\t\t\t\t\t\tstring \"%s\"\n", tmp_char_ptr );
      fprintf ( fp, "\t\t\t\t\t\tfontStyle FontStyle {\n" );
      fprintf ( fp, "\t\t\t\t\t\t\thorizontal FALSE \n" );
      fprintf ( fp, "\t\t\t\t\t\t}\n" );
      fprintf ( fp, "\t\t\t\t\t}\n" );
      fprintf ( fp, "\t\t\t\t}\n" );       // shape
      fprintf ( fp, "\t\t\t]\n" );       // end children
      fprintf ( fp, "\t}\n" );         // transform
   }
   
   // Adjustable elevation grid
   if ( g_data->appearance.elevation_grid.show ) {
      color = &g_data->appearance.elevation_grid.color;
      fprintf ( fp, "\tTransform {\n" );
      fprintf ( fp, "\t\ttranslation %f 0 -9.0\n", (float)min_x );
      fprintf ( fp, "\t\tchildren [\n" );
      fprintf ( fp, "\t\t\tShape {\n" );
      fprintf ( fp, "\t\t\t\tgeometry DEF GRID ElevationGrid {\n" );
      fprintf ( fp, "\t\t\t\t\txDimension 2\n");
      fprintf ( fp, "\t\t\t\t\tzDimension 2\n");
      fprintf ( fp, "\t\t\t\t\theight [0 0 0 0]\n");
      fprintf ( fp, "\t\t\t\t\txSpacing %d\n", (int)max_x );
      fprintf ( fp, "\t\t\t\t\tzSpacing %d\n", (int)(axis_z_frame-3) );
      fprintf ( fp, "\t\t\t\t}\n");        // grid
      fprintf ( fp, "\t\t\t\tappearance Appearance {\n");
      fprintf ( fp, "\t\t\t\t\tmaterial Material { \n" );
      fprintf ( fp, "\t\t\t\t\t\tdiffuseColor %1.2f %1.2f %1.2f\n", color->R, color->G, color->B);
      fprintf ( fp, "\t\t\t\t\t\ttransparency 0.4\n" );
      fprintf ( fp, "\t\t\t\t\t}\n" );
      fprintf ( fp, "\t\t\t\t}\n" );       // appearance
      fprintf ( fp, "\t\t\t}\n" );      // shape
      fprintf ( fp, "\t\t\tDEF PL PlaneSensor {\n" );
      fprintf ( fp, "\t\t\t\tminPosition 0 0\n");
      fprintf ( fp, "\t\t\t\tmaxPosition 100 100\n" );
      fprintf ( fp, "\t\t\t}\n" );         // PlaneSensor
      fprintf ( fp, "\t\t\tDEF MV_GRID Script {\n" );
      fprintf ( fp, "\t\t\t\teventIn SFVec3f dragPlane\n" );
      fprintf ( fp, "\t\t\t\teventOut MFFloat grHeight\n" );
      fprintf ( fp, "\t\t\t\tfield MFFloat h[0 0 0 0]\n\n"  );
      fprintf ( fp, "\t\t\t\turl \"javascript: \n" );
      fprintf ( fp, "\t\t\t\t\tfunction dragPlane(value) {\n" );
      fprintf ( fp, "\t\t\t\t\t\th[0] = value[1];\n" );
      fprintf ( fp, "\t\t\t\t\t\th[1] = value[1];\n" );
      fprintf ( fp, "\t\t\t\t\t\th[2] = value[1];\n" );
      fprintf ( fp, "\t\t\t\t\t\th[3] = value[1];\n" );
      fprintf ( fp, "\t\t\t\t\t\tgrHeight = h;\n" );
      fprintf ( fp, "\t\t\t\t\t} \" \n" );     // end function
      fprintf ( fp, "\t\t\t}\n" );       // Script
      fprintf ( fp, "\t\t]\n" );        // children
      // Routes
      fprintf ( fp, "ROUTE PL.translation_changed TO MV_GRID.dragPlane\n" );
      fprintf ( fp, "ROUTE MV_GRID.grHeight TO GRID.set_height\n" );
      fprintf ( fp, "\t}\n" );        // parent transform
   }
   
}

/*****************  graph_box_data  ******************/
int
graph_box_data (BoxGraphData *g_data, DATA_ACCESS_FNXS *fnxs, FILE *fp )
{
   float max_y;
   float min_y;
   float scale_y;
   float max_z;
   int   box_width, box_depth;
   int   n_tuples;
   int   i;
   int   rc;
   

   box_width = 10;
   box_depth = 10;
   max_z = MINFLOAT;
   
   if ( (rc = get_min_max_y( fnxs, g_data, &min_y, &max_y )) != 0)
      return rc;
   
   if ( (rc = get_x_z_field_axis ( fnxs, g_data, box_width, box_depth )) != 0)
      return rc;
   
   // Constrain graph to 100 max Y value
   scale_y = (float)(fabs(max_y - min_y) / 100.0);


   
   n_tuples = (*fnxs->num_tuples_fnx)(fnxs->instance_data);

   for ( i=0; i < n_tuples; i++ ) {
      graph_item_box (fp, fnxs, g_data, i, scale_y, min_y, box_width, box_depth, &max_z );
   }

   add_frame( fp, fnxs, g_data, min_y, max_y, max_z, box_width );
   
   return 0;
   
}

/***************  init_field_orders  ***************/
void
init_field_orders( DATA_ACCESS_FNXS *fnxs, BoxGraphData *g_data)
{
   const int f = 3;
   int       i, j;
   int       n_fields;
   
   DBFieldOrder **fields =
      (DBFieldOrder**)malloc(sizeof(DBFieldOrder*) * f);

   n_fields = (*fnxs->num_fields_fnx)( fnxs->instance_data );
   for ( i=0; i < 3; i++ ) {
      fields[i] = (DBFieldOrder*)malloc(sizeof(DBFieldOrder));
      fields[i]->axis   = NULL;
      fields[i]->n_axis = 0;
      fields[i]->field_name = NULL;
   }
   for ( j=0; j < n_fields; j++ ) {
      if ( strcasecmp( (*fnxs->field_name_fnx) (fnxs->instance_data, j), g_data->box_map->x_axis->column_name) == 0) {
         fields[X_FIELD]->field_name = strdup( g_data->box_map->x_axis->column_name );
         fields[X_FIELD]->field_order = j;
      }
      else if ( strcasecmp( (*fnxs->field_name_fnx) (fnxs->instance_data, j), g_data->box_map->y_axis->column_name) == 0) {
         fields[Y_FIELD]->field_name = strdup( g_data->box_map->y_axis->column_name );
         fields[Y_FIELD]->field_order = j;
      }
      else if ( strcasecmp( (*fnxs->field_name_fnx)(fnxs->instance_data, j), g_data->box_map->z_axis->column_name) == 0) {
         fields[Z_FIELD]->field_name = strdup( g_data->box_map->z_axis->column_name );
         fields[Z_FIELD]->field_order = j;
      }
   }            // for fields in relation

#ifdef DEBUG
   for ( i=0; i < 3; i++ ) {
      assert ( fields[i]->field_name != NULL );
   }
#endif
   
   g_data->fields = fields;
}

/****************  destroy_field_orders ***************/
void
destroy_field_orders (BoxGraphData *g_data )
{
   int       i;
   int       i_loc;
   DBFieldLocation *loc;
   
   DBFieldOrder **fields = g_data->fields;

   if ( fields != NULL ) {
      for (i=0; i<3; i++ ) {
         if ( fields[i] ) {
            if ( fields[i]->field_name )
               free (fields[i]->field_name);
            if ( fields[i]->n_axis > 0 ) {
               for (i_loc = 0; i_loc < fields[i]->n_axis; i_loc ++ ) {
                  loc = fields[i]->axis[i_loc];
                  DBFieldLocation_delete (loc);
               }
               free (fields[i]->axis);
            }
            
            free (fields[i]);
         }
      }
      free (fields);
   }
}

/****************  graph_box  *********************/
int
graph_box (BoxGraphData *g_data, DATA_ACCESS_FNXS *fnxs, const char* file_name )
{
   FILE  *fp;
   int    rc;


   if ( (rc = (*fnxs->initialize_fnx)(fnxs->instance_data)) != 0 )
      return rc;

   init_field_orders( fnxs, g_data );
   
   if ( (fp = fopen( file_name, "wb" )) == NULL )
      return 2;

   
   fprintf (fp, "#VRML V2.0 utf8\n");
   fprintf (fp, "Transform {\n");
   fprintf (fp, "children [\n");
   fprintf (fp, "\tViewpoint {\n");
   fprintf (fp, "\t\tdescription \"Start View\"\n" );
   fprintf (fp, "\t\tposition 0 50 300\n" );
   fprintf (fp, "\t}\n" );

   rc = graph_box_data( g_data, fnxs, fp);

   fprintf (fp, "]}\n");
   fclose  (fp);
   
   destroy_field_orders( g_data );
   (*fnxs->close_fnx)(fnxs->instance_data);
   return  rc;
}
