
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "../../boxmap.h"
#include "vrml-postgres.h"
#include "../../graph_types.h"

#ifdef POSTGRES_INTERFACE

/**********  exit_postgres  *****************/
void
exit_postgres( Postgres_Connection_Data *pg, int status_code )
{

   assert( pg != NULL );

   
   PQfinish( pg->conn );
   exit(status_code);
}


/**************  postgres_init  **********************/
int
postgres_init( DB_Instance_Data *instance )
{
   char     tmp_str[512];
   
   Postgres_Connection_Data *pg = (Postgres_Connection_Data*)malloc( sizeof(Postgres_Connection_Data) );
   assert(pg);
   if ( pg == NULL ) {
      fprintf ( stderr, "\nOut of memory allocating postgress initialization structures" );
      exit( 1 );
   }

   pg->conn = PQsetdb(instance->init_data->host_name,
                      instance->init_data->port,
                      instance->init_data->server_options,
                      NULL,
                      instance->init_data->data_base);
   if ( PQstatus (pg->conn) == CONNECTION_BAD ) {
      fprintf(stderr, "\nConnection to database '%s' failed.\n", instance->init_data->data_base );
      fprintf(stderr, "%s", PQerrorMessage(pg->conn));
      exit_postgres(pg, 1);
   }

   /* start a transaction block */
   pg->cursor = PQexec(pg->conn, "BEGIN");
   if (PQresultStatus(pg->cursor) != PGRES_COMMAND_OK) {
         fprintf (stderr, "\nBEGIN command failed\n");
         PQclear (pg->cursor);
         exit_postgres (pg, 1);
   }

   PQclear(pg->cursor);

   /*
     * fetch instances from the pg_database, the system catalog of
     * databases
     */
   snprintf( tmp_str, 512, "DECLARE mycursor CURSOR FOR select * from %s", instance->init_data->relation);
   pg->cursor = PQexec(pg->conn, tmp_str );
   if (PQresultStatus(pg->cursor) != PGRES_COMMAND_OK)  {
      fprintf (stderr, "DECLARE CURSOR command failed\n");
      PQclear (pg->cursor);
      exit_postgres (pg, 1);
   }
   PQclear(pg->cursor);
   
   pg->cursor = PQexec(pg->conn, "FETCH ALL in mycursor");
   if (PQresultStatus(pg->cursor) != PGRES_TUPLES_OK)  {
      fprintf (stderr, "FETCH ALL command didn't return tuples properly\n");
      PQclear (pg->cursor);
      exit_postgres (pg, 1);
   }

   instance->connection_information = pg;

   return 0;
   
}

/**************  postgres_close  *******************/
void
postgres_close ( DB_Instance_Data *instance)
{
   Postgres_Connection_Data *pg = (Postgres_Connection_Data*)instance->connection_information;

   if ( pg != NULL ) {
      if ( pg->cursor != NULL )
         PQclear (pg->cursor);
      PQfinish (pg->conn);
      free (pg);
   }
   instance->connection_information = NULL;
}

/*************  postgres_num_fields  ***************/
int
postgres_num_fields (DB_Instance_Data  *instance )
{
   Postgres_Connection_Data *pg = (Postgres_Connection_Data*)instance->connection_information;

   assert( pg != NULL );
   assert( pg->cursor != NULL );
   
   return PQnfields( pg->cursor );
}

/*************   postgres_field_name  ***************/
char*
postgres_field_name (DB_Instance_Data *instance, int field_number )
{
   Postgres_Connection_Data *pg = (Postgres_Connection_Data*)instance->connection_information;

   assert( pg != NULL );
   assert( pg->cursor != NULL );

   return PQfname (pg->cursor, field_number );
}

/*************   postgres_num_tuples  ***************/
int
postgres_num_tuples (DB_Instance_Data *instance )
{
   Postgres_Connection_Data *pg = (Postgres_Connection_Data*)instance->connection_information;

   assert( pg != NULL );
   assert( pg->cursor != NULL );

   return PQntuples (pg->cursor);
}

/*************   postgres_get_value  ***************/
char*
postgres_get_value (DB_Instance_Data *instance, int tuple, int field_number )
{
   Postgres_Connection_Data *pg = (Postgres_Connection_Data*)instance->connection_information;

   assert( pg != NULL );
   assert( pg->cursor != NULL );

   return PQgetvalue (pg->cursor, tuple, field_number );
   
}

/*************   postgres_is_null  ***************/
int
postgres_is_null (DB_Instance_Data *instance, int tuple, int field_number )
{
   Postgres_Connection_Data *pg = (Postgres_Connection_Data*)instance->connection_information;

   assert( pg != NULL );
   assert( pg->cursor != NULL );

   return  PQgetisnull (pg->cursor, tuple, field_number);
}

#else
void no_interface()
{
   fprintf ( stderr, "No Postgres interface. Define POSTGRES_INTERFACE\n" );
   /* Must define POSTGRES_INTERFACE */
   assert (0);
   exit(1);
}

int         postgres_init       ( DB_Instance_Data *d)  {no_interface(); return 0; }

void        postgres_close      ( DB_Instance_Data *d)  {no_interface();}

int         postgres_num_fields ( DB_Instance_Data *d)  {no_interface(); return 0; }
int         postgres_num_tuples ( DB_Instance_Data *d)  {no_interface(); return 0; }
char*       postgres_field_name ( DB_Instance_Data *d, int field_number)   {no_interface(); return NULL;}
char*       postgres_get_value  ( DB_Instance_Data *d, int tuple, int field_number )  {no_interface(); return NULL;}
int         postgres_is_null    ( DB_Instance_Data *d, int tuple, int field_number )  {no_interface(); return 0; }

#endif

