
#include "vrml-text.h"
#include <assert.h>
#include <string.h>
#include <stdio.h>

/**************  Text_Field_Tuple_new  ***************/
Text_Field_Tuple*
Text_Field_Tuple_new ()
{
   int i;
   Text_Field_Tuple *t = (Text_Field_Tuple*)malloc( sizeof (Text_Field_Tuple) );
   assert (t);

   for ( i=0; i < MAX_NUM_FIELDS; i++ ) {
      t->value[i] = NULL;
   }
   
   return t;
}

/****************  Text_Field_Tuple_delete  ***************/
void
Text_Field_Tuple_delete  (Text_Field_Tuple *t )
{
   int i;
   
   assert (t);
   if ( t ) {
      for (i=0; i < MAX_NUM_FIELDS; i++ )
         if ( t->value[i] )
            free (t->value[i]);
      free (t);
   }
   
}

/**************  Text_Field_Tuple_add_value  *****************/
void
Text_Field_Tuple_add_value (Text_Field_Tuple *t,
                            int              field_num,
                            const char       *value )
{
   assert (t);
   if ( field_num >= MAX_NUM_FIELDS ) {
      fprintf ( stderr, "\nCan not set field %d.  Maximum %d fields", field_num, MAX_NUM_FIELDS );
      return;
   }
   
   if ( t ) {
      if ( t->value[field_num] )
         free (t->value[field_num]);
      t->value[field_num] = strdup(value);
   }
   
}

/***********  Text_File_Connection_Data_new  *****************/
Text_File_Connection_Data*
Text_File_Connection_Data_new ()
{
   Text_File_Connection_Data  *d;
   int                        i;
   
   d = (Text_File_Connection_Data*)malloc( sizeof(Text_File_Connection_Data) );
   assert (d);

   for ( i=0; i < MAX_NUM_FIELDS; i++ ) {
      d->field_lookup[i].name = NULL;
   }
   d->num_fields = 0;

   d->tuples = NULL;
   d->num_tuples = 0;

   d->read_fp = NULL;
   
   return d;
}

/*************  Text_File_Connection_Data_delete  ****************/
void
Text_File_Connection_Data_delete (Text_File_Connection_Data *d)
{
   int i;

   for ( i=0; i < MAX_NUM_FIELDS; i++ ) {
      if ( d->field_lookup[i].name )
         free (d->field_lookup[i].name);
   }

   if ( d->tuples ) {
      for ( i=0; i < d->num_tuples; i++ ) {
         free (d->tuples[i]);
      }
      free (d->tuples);
   }

   if ( d->read_fp ) {
      fclose (d->read_fp);
      d->read_fp = NULL;
   }
   free (d);
}


/***************  Text_File_Connection_Data_add_tuple  ***************/
void
Text_File_Connection_Data_add_tuple ( Text_File_Connection_Data *d, Text_Field_Tuple *t)
{
   void *new_list;
   int  new_count;

   new_count = d->num_tuples + 1;
   new_list = realloc( d->tuples, sizeof(Text_Field_Tuple*) * new_count );

   if ( new_list ) {
      d->tuples = (Text_Field_Tuple**)new_list;
      d->tuples[ d->num_tuples ] = t;
      d->num_tuples ++;
   }
   
}


/**********************************************
 Data access functions
*********************************************/

/***********  text_file_init  ***************/
int
text_file_init (DB_Instance_Data *instance )
{
   char              *line;
   char              tmp_char[256];
   char              c;
   size_t            line_len, i_line, i_char;
   ssize_t           read_bytes;
   int               field_num;
   Text_Field_Tuple  *tuple;
   
   Text_File_Connection_Data *conn_data = Text_File_Connection_Data_new();

   conn_data->read_fp = fopen( instance->init_data->data_base, "rb" );
   if ( conn_data->read_fp == NULL ) {
      fprintf ( stderr, "\nCould not open data file %s\n", instance->init_data->data_base );
      exit(1);
   }

   // Now try to read the data
   line = malloc(line_len = 128);

   
   // First must come the column headers (field names)
   read_bytes = getline ( &line, &line_len, conn_data->read_fp );

   
   for ( i_line = 0; i_line < read_bytes;  ) {
      for ( i_char=0; i_char < 256 && i_line < read_bytes; ) {
         c = line[i_line++];
         //         printf ( "\n%s", (line+i_line));
         if ( (c == '\t' || c == '\n') && conn_data->num_fields < MAX_NUM_FIELDS ) {
            // Remove all leading tabs/spaces
            while ( (line[i_line] == '\t' || line[i_line] == ' ') && i_line < read_bytes )
               i_line ++;
            
            tmp_char[ i_char ] = 0;
            conn_data->field_lookup[ conn_data->num_fields++ ].name = strdup(tmp_char);
            break;
         }
         else {
            tmp_char[ i_char++ ] = c;
         }
      }
   }


   
   // Now try to read all data in rows
   while ( !ferror(conn_data->read_fp) && !feof( conn_data->read_fp ) ) {

      
      read_bytes = getline( &line, &line_len, conn_data->read_fp );
      
      if ( read_bytes != -1 ) {
         field_num = 0;
         tuple = Text_Field_Tuple_new();

      
         // Get all the fields in this row/tuple
         for ( i_line = 0; i_line < read_bytes;  ) {
            for ( i_char=0; i_char < 256 && i_line < read_bytes; ) {
               c = line[i_line++];
               if ( c == '\t' || c == '\n' ) {
                  // Remove all tabs
                  while ( (line[i_line] == '\t' || line[i_line] == ' ') && i_line < read_bytes )
                     i_line ++;
               
                  tmp_char[ i_char ] = 0;
                  Text_Field_Tuple_add_value (tuple, field_num++, tmp_char );
                  break;
               }
               else {
                  tmp_char[ i_char++ ] = c;
               }
            }
         }
         Text_File_Connection_Data_add_tuple (conn_data, tuple);
      }
   }
   
   free (line);


   instance->connection_information = conn_data;
   return 0;
}


                                             
/**************  text_file_close  *******************/
void
text_file_close ( DB_Instance_Data *instance)
{
   Text_File_Connection_Data *cd;
   cd = (Text_File_Connection_Data*)instance->connection_information;

   Text_File_Connection_Data_delete (cd);
   instance->connection_information = NULL;
}

/*************  text_file_num_fields  ***************/
int
text_file_num_fields (DB_Instance_Data  *instance )
{
   Text_File_Connection_Data *cd;
   cd = (Text_File_Connection_Data*)instance->connection_information;

   return cd->num_fields;
}

/*************   text_file_field_name  ***************/
char*
text_file_field_name (DB_Instance_Data *instance, int field_number )
{
   Text_File_Connection_Data *cd;
   cd = (Text_File_Connection_Data*)instance->connection_information;

   assert (field_number < cd->num_fields );
   if ( field_number < cd->num_fields )
      return cd->field_lookup[field_number].name;
   
   return NULL;
}

/*************   text_file_num_tuples  ***************/
int
text_file_num_tuples (DB_Instance_Data *instance )
{
   Text_File_Connection_Data *cd;
   cd = (Text_File_Connection_Data*)instance->connection_information;

   return cd->num_tuples;
}

/*************   text_file_get_value  ***************/
char*
text_file_get_value (DB_Instance_Data *instance, int tuple_number, int field_number )
{
   Text_File_Connection_Data *cd;
   Text_Field_Tuple          *tuple;
   
   cd = (Text_File_Connection_Data*)instance->connection_information;

   assert (field_number < cd->num_fields && tuple_number < cd->num_tuples );

   if ( field_number < cd->num_fields && tuple_number < cd->num_tuples ) {
      tuple = cd->tuples[tuple_number];
      return tuple->value[ field_number ];
   }
   return NULL;
}


/*************   text_file_is_null  ***************/
int
text_file_is_null (DB_Instance_Data *instance, int tuple_number, int field_number )
{
   Text_File_Connection_Data *cd;
   cd = (Text_File_Connection_Data*)instance->connection_information;

   return (text_file_get_value (instance, tuple_number, field_number) == NULL);
}


