/****************************************************************
 * See file COPYING included in distribution for license details
 ****************************************************************/
#include <stdlib.h>
#include "example.h"
#include "boxmap.h"
#include "db-interfaces/postgres/vrml-postgres.h"
#include "db-interfaces/text-file/vrml-text.h"
#include <unistd.h>
#include <string.h>

extern int  optind;
extern char *optarg;

/******************  usage  ********************/
void
usage( char* exe )
{
   printf ( "Usage: %s \n", exe );
   printf ( "\t -b backend [postgres textfile]\n" );
   printf ( "\t -x x_field\n" );
   printf ( "\t -y y_field\n" );
   printf ( "\t -z z_field\n" );
   printf ( "\t -o output_file\n" );
   printf ( "\n\t Optional parameters (appearance):\n" );
   printf ( "\t\t -1 \"r g b\" box-color\n" );
   printf ( "\t\t -2 \"r g b\" line color\n" );
   printf ( "\t\t -X  [0 1] print x-axis-heading\n" );
   printf ( "\t\t -3 \"r g b\" x-label color\n" );
   printf ( "\t\t -Z  [0 1] print z-axis-heading\n" );
   printf ( "\t\t -4 \"r g b\" z-label color\n" );
   printf ( "\t\t -E  [0 1] use adjustable elevation grid\n" );
   printf ( "\t\t -5 \"r g b\" elevation grid color\n" );
   printf ( "\t\t -j  [f e] 1=justify text with Fontstyle/Estimate\n" );
   printf ( "\t\t           Compensates(?) for varying support among viewers\n" );
   printf ( "\n\t Optional parameters (data-base):\n" );
   printf ( "\t\t -d data_base_name\n" );
   printf ( "\t\t -r relation\n" );
   printf ( "\t\t -h host_name\n" );
   printf ( "\t\t -p port\n" );
   
   exit(1);
}

/*****************  set_custom_color  ****************/
void
set_custom_color (SFColor *color, const char* rgb_str )
{
   float r, g, b;
   float v;
   char  tmp[32];
   char  c;
   int   i, j;
   int   len = strlen(rgb_str);

   r = g = b = -1;

   j = 0;
   // remove leading spaces
   while ( rgb_str[j] == ' ' )
      j++;
   
   for ( ; j < len; j++ ) {
      for ( i=0; j < len ; j++, i++  ) {
         c = rgb_str[j];
         tmp[i] = c;
         if ( c == ' ' ) {
            // space so try to construct number
            v = atof( tmp );
            if ( r < 0 )
               r = v;
            else if ( g < 0 )
               g = v;
            else if ( b < 0 )
               b = v;
            break;
         }
      }
   }
   
   if ( b < 0 )
      b = atof(tmp);

   
   SFColor_set( color, r, g, b );
}

/*************  main  ****************/
int
main (int argc, char* argv[] )
{
   BoxGraphData     *data;   
   BoxMap           *box_map;
   CoordMap         *coord_map;
   DATA_ACCESS_FNXS data_functions;
   char*            b_arg = NULL;
   char*            x_arg = NULL;
   char*            y_arg = NULL;
   char*            z_arg = NULL;
   char*            o_arg = NULL;
   char*            d_arg = NULL;
   char*            r_arg = NULL;
   char*            h_arg = NULL;
   char*            p_arg = NULL;
   int              arg;
   int              parsing_args = 1;
   int              ok;
   
   if ( argc < 2 )
      usage(argv[0]);

   /* Create our BoxGraphData structure to hold everything */
   data      = BoxGraphData_new ();


   /* Get the user parameters */
   while ( parsing_args ) {
      arg = getopt( argc, argv, "b:x:y:z:o:d:r:h:p:1:2:X:3:Z:4:E:5:j:" );
      //      printf ( "\n%c - %s", arg, optarg );
      switch( arg ) {
      case 'b':
         b_arg = strdup(optarg);
         break;
      case 'x':
         x_arg = strdup(optarg);
         break;
      case 'y':
         y_arg = strdup(optarg);
         break;
      case 'z':
         z_arg = strdup(optarg);
         break;
      case 'o':
         o_arg = strdup(optarg);
         break;
      case 'd':
         d_arg = strdup(optarg);
         break;
      case 'r':
         r_arg = strdup(optarg);
         break;
      case 'h':
         h_arg = strdup(optarg);
         break;
      case 'p':
         p_arg = strdup(optarg);
         break;
      case '1':
         /* Change color from default */
         set_custom_color( &data->appearance.box_color, optarg );
         break;
      case '2':
         /* Change color from default */
         set_custom_color( &data->appearance.graph_lines.color, optarg );
         break;
      case 'X':
         /* Choose to hide/show x_axis_heading */
         data->appearance.x_axis_heading.show = atoi(optarg);
         break;
      case '3':
         /* Change color from default */
         set_custom_color( &data->appearance.x_axis_labels.color, optarg );
         break;
      case 'Z':
         /* Choose to hide/show z_axis_heading */
         data->appearance.z_axis_heading.show = atoi(optarg);
         break;
      case '4':
         /* Change color from default */
         set_custom_color( &data->appearance.z_axis_labels.color, optarg );
         break;
      case 'E':
         /* Choose to hide/show elevation_grid */
         data->appearance.elevation_grid.show = atoi(optarg);
         break;
      case '5':
         /* Change color from default */
         set_custom_color( &data->appearance.elevation_grid.color, optarg );
         break;
      case 'j':
         /*
           Set the text_justify algorithm to compensate for variances in
           VRML browers
         */
         if ( optarg[0] == 'f' || optarg[0] == 'F' ) {
            data->appearance.text_justify_mode = TEXT_JUSTIFY_FONTSTYLE;
         }
         else if ( optarg[0] == 'e' || optarg[0] == 'E' ) {
            data->appearance.text_justify_mode = TEXT_JUSTIFY_ESTIMATE_1;
         }
         break;
      default:
         parsing_args = 0;
         break;
      }
   }

   /* Not enough info to continue */
   if ( b_arg == NULL || x_arg == NULL || y_arg == NULL || z_arg == NULL || o_arg == NULL ) {
      BoxGraphData_delete (data);
      usage(argv[0]);
   }
   

   /*
     Create our mapping object which will tell libvrmldb
     which fields will be mapped to which axis in the
     VRML world
   */
   box_map   = BoxMap_new ();

   /*
     Replace coordinates, defaults are always wrong!
   */

   /*
     Which field will be X axis and how should we
     display its name in the VRML world?
   */
   coord_map = CoordMap_new (x_arg, x_arg);
   BoxMap_replace_CoordMap_X( box_map, coord_map );

   /*
     Which field will be Z axis and how should we
     display its name in the VRML world?
   */
   coord_map = CoordMap_new (z_arg, z_arg);
   BoxMap_replace_CoordMap_Z( box_map, coord_map );

   
   /*
     Which field will be Y axis and how should we
     display its name in the VRML world?
   */
   coord_map = CoordMap_new (y_arg, y_arg);
   BoxMap_replace_CoordMap_Y( box_map, coord_map );

   /*  Replace BoxMap with one we defined */
   BoxGraphData_replace_BoxMap( data, box_map );

   /*
     Create DATA_ACCESS_FNXS and point to routines for
     desired backend
   */
   data_functions.instance_data = DB_Instance_Data_new();

   /*
     Set shallow ptr in DB_Instance_Data so we don't need to
     worry about deleting it ourselves.  DB_Instance_Data_delete
     will take care of it.
   */
   DB_Instance_Data_use_shallow_ptr( data_functions.instance_data,
                                     DB_Init_Data_new( d_arg,
                                                       r_arg,
                                                       h_arg,
                                                       p_arg,
                                                       NULL ));

   /*
     Set appropriate access functions for our backend
   */
   ok = 1;
   if ( strcmp( b_arg, "postgres" ) == 0 ) {
      /* Use Postgres */
      DATA_ACCESS_FNXS_set_functions  ( &data_functions,
                                        postgres_init,
                                        postgres_close,
                                        postgres_num_fields,
                                        postgres_num_tuples,
                                        postgres_field_name,
                                        postgres_get_value,
                                        postgres_is_null );
   }
   else if ( strcmp( b_arg, "textfile") == 0 ) {
      /* Use Text-File */
      DATA_ACCESS_FNXS_set_functions  ( &data_functions,
                                        text_file_init,
                                        text_file_close,
                                        text_file_num_fields,
                                        text_file_num_tuples,
                                        text_file_field_name,
                                        text_file_get_value,
                                        text_file_is_null );
   }
   else {
      /* Unknown backend */
      fprintf( stderr, "\nUnknown backend, %s", b_arg );
      ok = 0;
   }
   
   
   if ( ok ) {
      /* Create VRML graph in /tmp/city.wrl */
      graph_box (data, &data_functions, "/tmp/city.wrl");
   

      /* Delete our allocated structs */
      BoxGraphData_delete     (data);
      DB_Instance_Data_delete (data_functions.instance_data);
   }

   /* Delete any ptrs to user parameters */
   if ( b_arg )
      free (b_arg);
   if ( x_arg )
      free (x_arg);
   if ( y_arg )
      free (z_arg);
   if ( x_arg )
      free (o_arg);
   if ( x_arg )
      free (d_arg);
   if ( x_arg )
      free (r_arg);
   if ( x_arg )
      free (h_arg);
   if ( x_arg )
      free (p_arg);

   if ( !ok )
      usage( argv[0] );

   /* Out of here */
   exit(0);

   /* Keep GCC happy */
   return 0;
}
